


def print_bytes(byte_collection, delimiter=" "):
    """
    Output a human readable string (with spaces) of a byte list

    >>> print_bytes(b'\x02\x65)
    '02 65'

    >>> print_bytes(b'\xff\xaa', delimiter"-")
    'ff-aa'
    """

    byte_list = [bytes([b]).hex() for b in byte_collection]
    bytes_str = delimiter.join(byte_list)

    return bytes_str



import collections
import enum
import io
import sys
import serial
import serial.tools.list_ports
import time
import struct

from roktools import logger

from . import helpers

TIMEOUT_DEF = 5

def compute_checksum(message_without_synch_nor_checksum):
    """
    Compute the checksum from a message (without synch nor checksum)

    >>> MESSAGE = PREAMBLE + PacketType.RXM_RAWX.value + bytes.fromhex('10 00 00 00 00 00 00 c8 ba 40 00 00 12 00 00 01 5d a1 fa bd')
    >>> checksum = compute_checksum(MESSAGE[2:-2])
    >>> bytes(checksum).hex()
    'fabd'
    """

    CK_A = 0
    CK_B = 0

    for b in message_without_synch_nor_checksum:
        CK_A = (CK_A + b) & 0xff
        CK_B = CK_B + CK_A & 0xff
    
    return CK_A, CK_B

def validate_checksum(packet_without_sync):
    """

    >>> MESSAGE = PREAMBLE + PacketType.RXM_RAWX.value + bytes.fromhex('10 00 00 00 00 00 00 c8 ba 40 00 00 12 00 00 01 5d a1 fa bd')
    >>> validate_checksum(MESSAGE[2:])
    True
    """

    ck_a, ck_b = compute_checksum(packet_without_sync[:-2])
    ck_a_ref = packet_without_sync[-2]
    ck_b_ref = packet_without_sync[-1]

    is_valid = (ck_a == ck_a_ref and ck_b == ck_b_ref)
    if not is_valid:
        msg_str = helpers.print_bytes(packet_without_sync)
        logger.critical('Invalid checksum for packet [ {} ] (nbytes {}), expected [ {} ], got [ {} ]'.format(msg_str, len(packet_without_sync), bytes([ck_a_ref, ck_b_ref]), bytes([ck_a, ck_b])))

    return is_valid


def ublox_packets(stream, timeout):

    while(stream.is_open):
        __go_to_next_packet__(stream, timeout)
        packet = get_next_package(stream)
        if packet:
            yield packet
        else:
            sys.stderr.write("Invalid checksum\n")


def get_next_package(stream):

    packet_type = stream.read(2)
    length = stream.read(2)

    payload = stream.read(int.from_bytes(length, "little"))

    ck_a = stream.read(1)
    ck_b = stream.read(1)

    packet = packet_type + length + payload + ck_a + ck_b

    valid_checksum = validate_checksum(packet)

    return packet if valid_checksum else None


def parse(packet_without_sync):

    packet_type = PacketType(packet_without_sync[0:2])

    parser_name = 'parse_' + packet_type.name.lower()
    parser = globals()[parser_name]

    fields = parser(packet_without_sync)
    
    return packet_type, fields


def extract_type_bytes(packet_without_sync):
    return packet_without_sync[0:2].hex()



def get_n_bytes_of_config_value(key_id):
    """
    Get the storage size of the configuration value from its KeyID, as per 
    description of section 7.2 of ublox interface description

    >>> get_n_bytes_of_config_value(KeyId.CFG_MSGOUT_UBX_RXM_RAWX_USB)
    1
    >>> get_n_bytes_of_config_value(KeyId.CFG_UART2_BAUDRATE)
    4
    >>> get_n_bytes_of_config_value(KeyId.CFG_RATE_MEAS)
    2
    """

    size = int((key_id.value >> 28) & 0x07)

    if size <= 2: return 1
    if size == 3: return 2
    if size == 4: return 4
    if size == 5: return 8


# -----

def send(packet_type, payload, serial_stream):
    """
    Send a payload of a given type to the ublox receiver
    """

    def build_packet(packet_type, payload):
    
        packet = struct.pack('<BBBBH', *PREAMBLE, *(packet_type.value), len(payload))
        packet += payload
        (ck_a, ck_b) = compute_checksum(packet[2:])
        packet += struct.pack('<BB', ck_a, ck_b)
     
        return packet

    packet = build_packet(packet_type, payload)

    bytes_str = helpers.print_bytes(packet)
    n_bytes_written = serial_stream.write(packet)
    logger.debug(f'Packet sent    [ {bytes_str} ], [ {n_bytes_written} ] bytes written')
    return n_bytes_written

# -----

def send_hex_string(string, serial_stream):
    """
    Send a serialized (stringified) hex string that contains both the packet
    type and payload. The method builds the rest of the packet and sends it to
    the receiver. The string expects the packet type, length and payload.

    An example to send CFG-MON-VER (poll request): "0A 04 00 00"
    """

    msg = bytes.fromhex(string)

    (ck_a, ck_b) = compute_checksum(msg)
    packet = struct.pack('<BB', *PREAMBLE) + msg + struct.pack('<BB', ck_a, ck_b)
    n_bytes_written = serial_stream.write(packet)

    logger.debug(f'Packet sent    [ {helpers.print_bytes(packet)} ], [ {n_bytes_written} ] bytes written')
    return n_bytes_written

# -----

def get_packet_length(packet_without_sync):
    return struct.unpack_from('<H', packet_without_sync, 2)[0]

class PacketTransaction:

    def __init__(self):
        self.packet_received = False
        self.ack_received = False

    def completed(self):
        return self.packet_received and self.ack_received


def submit_valget_packet(key_ids, serial_stream, timeout=TIMEOUT_DEF):

    PACKET_TYPE = PacketType.CFG_VALGET

    payload = struct.pack('<BBBB', 0, 0, 0, 0)

    for key_id in key_ids:
        payload += struct.pack('L', key_id.value)

    send(PACKET_TYPE, payload, serial_stream)

    packet_transaction = PacketTransaction()

    key_values = {}
    t_start = time.time()
    for packet in ublox_packets(serial_stream, timeout):

        elapsed_time = time.time() - t_start

        try:
            incoming_packet_type, parsed_packet = parse(packet)
        except KeyError:
            byte_str = helpers.print_bytes(packet)
            logger.debug('unsupported packet [ {} ... ]'.format(byte_str))
            continue

        logger.debug('packet detected    [ {} ]'.format(incoming_packet_type))

        if incoming_packet_type == PACKET_TYPE:
            cfg_data = parsed_packet.cfgData
            key_values = __extract_cfg_data__(cfg_data)
            logger.debug('detected [ {} ] configuration items'.format(len(key_values)))
            packet_transaction.packet_received = True
            if packet_transaction.completed():
                break

        elif incoming_packet_type == PacketType.ACK_ACK:
            if __ack_nck_corresponds_to_packet__(PACKET_TYPE, parsed_packet):
                logger.debug(f'ACK for [ {PACKET_TYPE} ] received!')
                packet_transaction.ack_received = True
                if packet_transaction.completed():
                    break

        elif incoming_packet_type == PacketType.ACK_NAK:
            if __ack_nck_corresponds_to_packet__(PACKET_TYPE, parsed_packet):
                logger.critical(f'NAK for [ {PACKET_TYPE} ] received!')
                break

        elif elapsed_time >= timeout:
            logger.critical('Time out reached without receiving the polled VALGET message')
            break

    return key_values

# ------------------------------------------------------------------------------

def submit_valset_packet(key_value_dict, serial_stream, timeout=TIMEOUT_DEF):
    """
    Set configuration items based on an input dictionary of key_ids and its
    corresponding value
    """

    PACKET_TYPE = PacketType.CFG_VALSET

    n_items_sent = 0

    sorted_dict = dict(sorted(key_value_dict.items(), key=lambda x : x[0].value))
    for i,(key_id,value) in enumerate(sorted_dict.items()):
        
        #n_transaction = 2 if i else 1
        n_transaction = 1
        payload = struct.pack('<BBBB', 1, 7, n_transaction, 0)

        logger.debug('submit_valset_packet::key_id [ {} ] --> [ {} ] (transaction {})'.format(key_id, value, n_transaction))
        try:

            payload += struct.pack('<L', key_id.value)
            fmt = KEY_ID_VALUE_FORMATS.get(key_id, 'B')

            payload += struct.pack(fmt, value)
        except Exception as e:
            logger.warning(f'Error processing [ {key_id} ]: {str(e)}')
            continue

        send(PACKET_TYPE, payload, serial_stream)
        wait_for_ack_nak(PACKET_TYPE, serial_stream, timeout=TIMEOUT_DEF)

        n_items_sent += 1

    if n_items_sent:
        payload = struct.pack('<BBBB', 1, 7, 3, 0)
        send(PACKET_TYPE, payload, serial_stream)
        wait_for_ack_nak(PACKET_TYPE, serial_stream, timeout=TIMEOUT_DEF)

    logger.info('Ending transaction for VALSET!')



def wait_for_ack_nak(packet_type, serial_stream, timeout=TIMEOUT_DEF):

    t_start = time.time()
    for packet in ublox_packets(serial_stream, timeout):

        try:
            incoming_packet_type, parsed_packet = parse(packet)
        except KeyError:
            logger.debug('packet [ {} ] unsupported'.format(PacketType(packet[0:2])))
            continue

        logger.debug(f'Received packet {incoming_packet_type}')

        if incoming_packet_type == PacketType.ACK_ACK:
            if __ack_nck_corresponds_to_packet__(packet_type, parsed_packet):
                return True

        elif incoming_packet_type == PacketType.ACK_NAK:
            if __ack_nck_corresponds_to_packet__(packet_type, parsed_packet):
                raise RuntimeError(f'NAK for [ {packet_type} ] received!')

        elif time.time() - t_start >= timeout:
            raise TimeoutError(f'Time out reached without receiving a confirmation for {packet_type}')

    raise RuntimeError('End of input stream and no ACK/NCK received')

def wait_for_packet(packet_type, serial_stream, timeout=TIMEOUT_DEF):

    t_start = time.time()
    for packet in ublox_packets(serial_stream, timeout):

        try:
            incoming_packet_type, parsed_packet = parse(packet)
        except KeyError:
            logger.debug('packet [ {} ] unsupported'.format(PacketType(packet[0:2])))
            continue

        if incoming_packet_type == packet_type:
            return packet

        elif incoming_packet_type == PacketType.ACK_ACK:
            logger.info(f'ACK  for [ {packet_type} ] received!')

        elif incoming_packet_type == PacketType.ACK_NAK:
            if __ack_nck_corresponds_to_packet__(packet_type, parsed_packet):
                raise RuntimeError(f'NAK for [ {packet_type} ] received!')

        elif time.time() - t_start >= timeout:
            raise TimeoutError(f'Time out reached without receiving a confirmation for {packet_type}')

    raise RuntimeError(f'End of input stream and no packet {packet_type} received')




def get_string_from_bytearray(b):
    """
    >>> b =  bytes.fromhex("4d 4f 44 3d 5a 45 44 2d 46 39 50 00 00")
    >>> get_string_from_bytearray(b)
    'MOD=ZED-F9P'
    """
    s = str(b.decode('utf-8'))
    s = s[0:s.find('\x00')]
    return s

def get_receiver_info(serial_stream, timeout=5):

    send(PacketType.MON_VER, struct.pack('<'), serial_stream)

    for packet in ublox_packets(serial_stream, timeout):

        try:
            packet_type, parsed_packet = parse(packet)

            if packet_type == PacketType.MON_VER:

                res = {
                    'sw_version' : get_string_from_bytearray(parsed_packet.swVersion),
                    'hw_version' : get_string_from_bytearray(parsed_packet.hwVersion),
                }

                extension_bytearray = parsed_packet.extension
                for l in range(int(len(extension_bytearray) / 30)):
    
                    fields = get_string_from_bytearray(extension_bytearray[l*30:]).split('=')
                    if len(fields) == 2:
                        res[fields[0]] = fields[1]
                        
                return res

        except KeyError:
            pass


def submit_valset_from_valget_line(line, serial_stream):
    """
    Copy a CFG-VALGET line contents and use them to submit a CFG-VALSET 
    
    >>> line = 'MON-VER - 0A 04 DC 00 45 ...'
    >>> submit_valset_from_valget_line(line, None)
    False
    """
    SKIP_NBYTES = 8
    
    if not line.startswith('CFG-VALGET'):
        return False
    
    logger.debug(f'Sending contents of line [ {line} ]')
    packet = line.split(' - ')[1]
    payload_str = packet[SKIP_NBYTES*3:]
    
    payload = struct.pack('<BBBB', 0, 7, 0, 0)
    payload += bytes.fromhex(payload_str)
    
    n_bytes_sent = send(PacketType.CFG_VALSET, payload, serial_stream)
    wait_for_ack_nak(PacketType.CFG_VALSET, serial_stream)
    
    return n_bytes_sent
    


PAYLOAD_OFFSET=4

# Packet classes and parsers
Ack = collections.namedtuple('Ack', ['clsID', 'msgID'])
def parse_ack_ack(packet_without_sync):

    HEADER_FMT = '<BB'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return Ack(*fields)

def parse_ack_nak(packet_without_sync):
    return parse_ack_ack(packet_without_sync)

MonVer = collections.namedtuple('MonVer', ['swVersion', 'hwVersion', 'extension'])
def parse_mon_ver(packet_without_sync):

    length = get_packet_length(packet_without_sync)
    N = int((length - 40) / 30)
    HEADER_FMT = '<30s10s{}s'.format(N * 30)

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return MonVer(*fields)

MonHw = collections.namedtuple('MonHw', ['pinSel', 'pinBank', 'pinDir', 'pinVal', 
                                         'noisePerMS', 'agcCnt', 'aStatus', 'aPower', 
                                         'flags', 'reserved1', 'usedMask', 'VP', 
                                         'jamInd', 'reserved2', 'pinIrq', 'pullH', 'pullL'])
def parse_mon_hw(packet_without_sync):

    HEADER_FMT = '<4p4p4p4pHHBBpB4p17pB2p4p4p4p'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return MonHw(*fields)

RxmRawx = collections.namedtuple('RxmRawx', ['rcvTow', 'week', 'leapS', 'numMeas', 'recStat', 'version', 'reserved1'])
def parse_rxm_rawx(packet_without_sync):

    HEADER_FMT = '<dHbBBB2s'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return RxmRawx(*fields)


RxmSfrbx = collections.namedtuple('RxmSfrbx', ['gnssId', 'svId', 'reserved1', 'freqId', 'numWords', 'chn', 'version', 'reserved2'])
def parse_rxm_sfrbx(packet_without_sync):

    HEADER_FMT = '<BBBBBBBB'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return RxmSfrbx(*fields)


NavStatus = collections.namedtuple('NavStatus', ['iTOW', 'gpsFix', 'flags', 'fixStat', 'flags2', 'ttff', 'msss'])
def parse_nav_status(packet_without_sync):

    HEADER_FMT = '<LBpppLL'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return NavStatus(*fields)


NavPosecef = collections.namedtuple('NavPosecef', ['iTOW', 'ecefX', 'ecefY', 'ecefZ', 'pAcc'])
def parse_nav_posecef(packet_without_sync):

    HEADER_FMT = '<LlllL'

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return NavPosecef(*fields)


NavPosllh = collections.namedtuple('NavPosllh', ['iTOW', 'lon', 'lat', 'height', 'hMSL', 'hAcc', 'vAcc'])
def parse_nav_posllh(packet_without_sync):

    HEADER_FMT = '<LllllLL'

    fields = list(fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET))

    # Scale parameters
    fields[1] = fields[1] * 1.0e-7
    fields[2] = fields[2] * 1.0e-7

    return NavPosllh(*fields)

CfgValget = collections.namedtuple('CfgValget', ['version', 'layer', 'reserved1', 'cfgData'])
def parse_cfg_valget(packet_without_sync):

    N = get_packet_length(packet_without_sync) - 4
    HEADER_FMT = '<BB2s{}s'.format(N)

    fields = struct.unpack_from(HEADER_FMT, packet_without_sync, PAYLOAD_OFFSET)

    return CfgValget(*fields)


PREAMBLE = b'\xb5\x62'

CLASS_ID_NAV = b'\x01'
CLASS_ID_RXM = b'\x02'
CLASS_ID_ACK = b'\x05'
CLASS_ID_CFG = b'\x06'
CLASS_ID_MON = b'\x0A'

def make_packet_type(class_id, msg_id):
    return PacketType(bytes([class_id, msg_id]))


class PacketType(enum.Enum):

    NAV_POSECEF = CLASS_ID_NAV + b'\x01'
    NAV_POSLLH = CLASS_ID_NAV + b'\x02'
    NAV_STATUS = CLASS_ID_NAV + b'\x03'
    NAV_SVIN = CLASS_ID_NAV + b'\x3b'

    RXM_SFRBX = CLASS_ID_RXM + b'\x13'
    RXM_RAWX = CLASS_ID_RXM + b'\x15'
    
    ACK_ACK = CLASS_ID_ACK + b'\x01'
    ACK_NAK = CLASS_ID_ACK + b'\x00'
    
    CFG_CFG = CLASS_ID_CFG + b'\x09'
    CFG_VALSET = CLASS_ID_CFG + b'\x8a'
    CFG_VALGET = CLASS_ID_CFG + b'\x8b'

    MON_VER = CLASS_ID_MON + b'\x04'
    MON_HW = CLASS_ID_MON + b'\x09'

class KeyId(enum.Enum):

    CFG_CLOCK_OSC_FREQ = 0x40a4000d
    CFG_I2COUTPROT_UBX = 0x10720001

    CFG_INFMSG_UBX_USB = 0x20920004

    CFG_MSGOUT_UBX_RXM_RAWX_I2C = 0x209102a4
    CFG_MSGOUT_UBX_RXM_RAWX_UART1 = 0x209102a5
    CFG_MSGOUT_UBX_RXM_RAWX_UART2 = 0x209102a6
    CFG_MSGOUT_UBX_RXM_RAWX_USB = 0x209102a7
    CFG_MSGOUT_UBX_RXM_RAWX_SPI = 0x209102a8

    CFG_MSGOUT_UBX_NAV_POSECEF_USB = 0x20910027
    CFG_MSGOUT_UBX_RXM_SFRBX_USB = 0x20910234

    CFG_MSGOUT_NMEA_ID_GSA_I2C = 0x209100bf
    CFG_MSGOUT_NMEA_ID_GSA_UART1 = 0x209100c0
    CFG_MSGOUT_NMEA_ID_GSA_UART2 = 0x209100c1
    CFG_MSGOUT_NMEA_ID_GSA_USB = 0x209100c2
    CFG_MSGOUT_NMEA_ID_GSA_SPI = 0x209100c3

    CFG_RATE_MEAS = 0x30210001
    CFG_RATE_NAV = 0x30210002
    CFG_RATE_TIMEREF = 0x20210003

    CFG_UART1_BAUDRATE = 0x40520001
    CFG_UART1_STOPBITS = 0x20520002
    CFG_UART1_DATABITS = 0x20520003
    CFG_UART1_PARITY = 0x20520004

    CFG_UART2_BAUDRATE = 0x40530001
    CFG_UART2_STOPBITS = 0x20530002
    CFG_UART2_DATABITS = 0x20530003
    CFG_UART2_PARITY   = 0x20530004

    CFG_SIGNAL_GPS_L1CA_ENA  = 0x10310001
    CFG_SIGNAL_GPS_L2C_ENA   = 0x10310003
    CFG_SIGNAL_GAL_E1_ENA    = 0x10310007
    CFG_SIGNAL_GAL_E5B_ENA   = 0x1031000a
    CFG_SIGNAL_BDS_B1_ENA    = 0x1031000d
    CFG_SIGNAL_BDS_B2_ENA    = 0x1031000e
    CFG_SIGNAL_QZSS_L1CA_ENA = 0x10310012
    CFG_SIGNAL_QZSS_L2C_ENA  = 0x10310015
    CFG_SIGNAL_GLO_L1_ENA    = 0x10310018
    CFG_SIGNAL_GLO_L2_ENA    = 0x1031001a
    CFG_SIGNAL_GPS_ENA       = 0x1031001f
    CFG_SIGNAL_GAL_ENA       = 0x10310021
    CFG_SIGNAL_BDS_ENA       = 0x10310022
    CFG_SIGNAL_QZSS_ENA      = 0x10310024
    CFG_SIGNAL_GLO_ENA       = 0x10310025

    CFG_NAVSPG_USE_PPP = 0x10110019

KEY_ID_VALUE_FORMATS = {
    KeyId.CFG_RATE_MEAS : 'H',
    KeyId.CFG_RATE_NAV: 'H',
    KeyId.CFG_CLOCK_OSC_FREQ : 'L',
    KeyId.CFG_UART1_BAUDRATE : 'L',
    KeyId.CFG_UART2_BAUDRATE : 'L',
    KeyId.CFG_NAVSPG_USE_PPP : '?',
    KeyId.CFG_MSGOUT_UBX_RXM_RAWX_USB : '?',
    KeyId.CFG_MSGOUT_UBX_NAV_POSECEF_USB : '?',
    KeyId.CFG_MSGOUT_UBX_RXM_SFRBX_USB : '?',
    KeyId.CFG_SIGNAL_GPS_L1CA_ENA : '?',
    KeyId.CFG_SIGNAL_GPS_L2C_ENA : '?',
    KeyId.CFG_SIGNAL_GAL_E1_ENA : '?',
    KeyId.CFG_SIGNAL_GAL_E5B_ENA : '?',
    KeyId.CFG_SIGNAL_BDS_B1_ENA : '?',
    KeyId.CFG_SIGNAL_BDS_B2_ENA : '?',
    KeyId.CFG_SIGNAL_QZSS_L1CA_ENA : '?',
    KeyId.CFG_SIGNAL_QZSS_L2C_ENA : '?',
    KeyId.CFG_SIGNAL_GLO_L1_ENA : '?',
    KeyId.CFG_SIGNAL_GLO_L2_ENA : '?',
    KeyId.CFG_SIGNAL_GPS_ENA : '?',
    KeyId.CFG_SIGNAL_GAL_ENA : '?',
    KeyId.CFG_SIGNAL_BDS_ENA : '?',
    KeyId.CFG_SIGNAL_QZSS_ENA : '?',
    KeyId.CFG_SIGNAL_GLO_ENA : '?'
}

# ------------------------------------------------------------------------------

def parse_key_id_value(key_id, value):
    """
    Get the value from packed format
    
    >>> parse_key_id_value(KeyId.CFG_RATE_MEAS, bytes([0xe8, 0x03]))
    1000
    >>> parse_key_id_value(KeyId.CFG_SIGNAL_GPS_ENA, bytes([0x00]))
    0
    >>> parse_key_id_value(KeyId.CFG_SIGNAL_GPS_ENA, bytes([0x01]))
    1
    """

    # Check if value is iterable
    try:
        iter(value)
    except TypeError:
        value = [value] # need to convert to iterable

    fmt = KEY_ID_VALUE_FORMATS.get(key_id, 'B')

    bytes_value = bytes(value)
    parsed_value = struct.unpack_from('<{}'.format(fmt), bytes_value)[0]

    return parsed_value

# ------------------------------------------------------------------------------

def build_key_id_value(key_id, value):
    """
    Pack the value according to the key_id

    >>> build_key_id_value(KeyId.CFG_RATE_MEAS, 1000).hex()
    'e803'
    """
   
    fmt = '<{}'.format(KEY_ID_VALUE_FORMATS.get(key_id, 'B'))
    return struct.pack(fmt, value)

# ------------------------------------------------------------------------------

def __go_to_next_packet__(stream, timeout):
    """
    Skip bytes in the stream until a new preamble is found
    """

    byte_1 = None
    byte_2 = None

    t_start = time.time()

    while(True):

        elapsed_time = time.time() - t_start
        
        if byte_1 != PREAMBLE[0]:
            byte_1 = int.from_bytes(stream.read(1), "big")
            
        else:
            byte_2 = int.from_bytes(stream.read(1), "big")

            if byte_2 == PREAMBLE[1]:
                break
            if byte_2 == PREAMBLE[0]:
                byte_1 = byte_2

        if elapsed_time > timeout:
            raise TimeoutError('Unable to find a packet in the given time frame')

    return True

# ------------------------------------------------------------------------------

def __ack_nck_corresponds_to_packet__(expected_type, ack_parsed_packet):

    class_id = ack_parsed_packet.clsID
    msg_id = ack_parsed_packet.msgID

    incoming_type = make_packet_type(class_id, msg_id)

    return expected_type == incoming_type

# ------------------------------------------------------------------------------

def __extract_cfg_data__(cfg_data):

    byte_stream = io.BytesIO(cfg_data)

    key_values = {}

    while True:
        raw = byte_stream.read(4)
        if len(raw) == 0:
            break

        key_id = KeyId(struct.unpack_from('<L', raw)[0])
        N = get_n_bytes_of_config_value(key_id)

        raw = byte_stream.read(N)
        if len(raw) == 0:
            break

        fmt = '<{}s'.format(N)
        value = struct.unpack_from(fmt, raw)[0]
        key_values[key_id] = value

    byte_stream.close()

    return key_values



import collections
import json
import os.path
import struct
import serial
import time
import threading
import tempfile
import shutil
import queue

from roktools import logger

from . import core
from . import constants
from . import helpers

# ------------------------------------------------------------------------------

def get(serial_stream):
    """
    Build an output dictionary of configuration from the settings defined in the
    package
    """

    config = {}

    key_ids = list(KEY_IDS[constants.LOGGING_STR].values()) + \
                __get_key_ids_from_constellations_key_ids_dict__()

    try:
        settings = core.submit_valget_packet(key_ids, serial_stream)
        import copy
        config = copy.deepcopy(KEY_IDS)

        config_logging = config[constants.LOGGING_STR]
        for config_item in config_logging:
            key_id = config_logging[config_item]
            value = core.parse_key_id_value(key_id, settings[key_id])
            if key_id == core.KeyId.CFG_RATE_MEAS:
                    value = value / 1000
            config_logging[config_item] = value

        config_constellations = config[constants.CONSTELLATIONS_STR]
        for config_constellation in config_constellations.values():
            key_id = config_constellation[constants.ENABLED_STR]
            value = core.parse_key_id_value(key_id, settings[key_id])
            config_constellation[constants.ENABLED_STR] = value


            config_signals = config_constellation[constants.SIGNALS_STR]
            for signal_str in config_signals:
                key_id = config_signals[signal_str]
                value = core.parse_key_id_value(key_id, settings[key_id])
                config_signals[signal_str] = value

    except:
        pass

    return config

# ------------------------------------------------------------------------------

def set_from_ucenter_file(serial_stream, fh):

    res = False

    if fh.readable():
        for line in fh:
            submit_res = core.submit_valset_from_valget_line(line, serial_stream)
            if submit_res:
                logger.info('Applied config line [ {} ]'.format(line))
            else:
                logger.warning('Could not apply config line [ {} ]'.format(line))

            res = res or submit_res
            time.sleep(0.2)

    return res

# ------------------------------------------------------------------------------

def set_from_json_file(serial_stream, fh):

    res = False

    if fh.readable():
        res = set_from_dict(serial_stream, json.loads(fh.read()))
        time.sleep(0.2)

    return res

# ------------------------------------------------------------------------------

def set_from_dict(serial_stream, config_dict):

    logger.debug(f'Incoming configuration parameters {config_dict}')
    config = {}

    logging_dict = config_dict.get(constants.LOGGING_STR, {})
    config.update(__logging_dict_to_keyids__(logging_dict))

    constellation_dict = config_dict.get(constants.CONSTELLATIONS_STR, {})
    config.update(__constellations_dict_to_keyids__(constellation_dict))

    if config:
        logger.debug(f'Configuration to apply {config}')
        core.submit_valset_packet(config, serial_stream)
        return True
    
    else:
        logger.warning('No configuration has been applied')
        return False

# ------------------------------------------------------------------------------

def __get_key_ids_from_constellations_key_ids_dict__():

    out = []

    for d in KEY_IDS[constants.CONSTELLATIONS_STR].values():
        key_id = d[constants.ENABLED_STR]
        out.append(key_id)

        signal_dict = d[constants.SIGNALS_STR]
        for key_id in signal_dict.values():
            out.append(key_id)

    return out

# ------------------------------------------------------------------------------

def __logging_dict_to_keyids__(config_dict):

    smoothing = config_dict.get(constants.SMOOTHING_STR, False)
    rate = config_dict.get(constants.RATE_STR, 1)
    measurements = config_dict.get(constants.MEASUREMENTS_STR, True)
    ephemeris = config_dict.get(constants.EPHEMERIS_STR, True)
    solutions = config_dict.get(constants.SOLUTION_STR, True)

    out = {
        core.KeyId.CFG_NAVSPG_USE_PPP : smoothing,
        core.KeyId.CFG_RATE_MEAS : int(rate * 1000), # s -> ms
        core.KeyId.CFG_MSGOUT_UBX_RXM_RAWX_USB : measurements,
        #core.KeyId.CFG_RATE_NAV : rate,
        core.KeyId.CFG_MSGOUT_UBX_NAV_POSECEF_USB : solutions, 
        core.KeyId.CFG_MSGOUT_UBX_RXM_SFRBX_USB : ephemeris
    }

    return out

__logging_dict_to_keyids__.__doc__ = f""" 
    {{
        {constants.SMOOTHING_STR}: false,
        {constants.RATE_STR}: 30,
        {constants.MEASUREMENTS_STR}: true,
        {constants.EPHEMERIS_STR}: true,
        {constants.SOLUTION_STR}: true,
        {constants.FILE_ROTATION_STR}: 86400
    }}
"""

# ------------------------------------------------------------------------------

def __constellations_dict_to_keyids__(constellations_config):
    """
    Method to convert from JSON configuration dictionary to a flat dictionary
    of receiver key_ids and values
    
    :params config_dict: is the full configuration dictioary that includes,
    in particular, the "constellations" key.
    {
            "BDS": {
                "enabled": true,
                "signals": {"2I":true}
            },
            "GAL": {
                "enabled": true,
                "signals": {"7B":true,"1C":true}
            },
            "GLO": {
                "enabled": true,
                "signals": {"2C":true,"1C":true}
            },
            "GPS": {
                "enabled": true,
                "signals": {"2L":true,"1C":true}
            },
            "QZSS": {
                "enabled": true,
                "signals": {"2W":true,"1C":true}
            }
    }
    """

    out = {}

    constellations_keyids = KEY_IDS[constants.CONSTELLATIONS_STR]

    for constellation_str in constellations_keyids:

        constellation_keyids = constellations_keyids[constellation_str]
        constellation_config = constellations_config.get(constellation_str, {})

        key_id = constellation_keyids[constants.ENABLED_STR]
        enabled = constellation_config.get(constants.ENABLED_STR, False)
        
        out[key_id] = enabled

        signals_keyids = constellation_keyids[constants.SIGNALS_STR]
        signals_config = constellation_config.get(constants.SIGNALS_STR, {})

        for signal_str in signals_keyids:
            key_id = signals_keyids[signal_str]
            enabled = signals_config.get(signal_str, False)
            out[key_id] = enabled

    return out

# ------------------------------------------------------------------------------

KEY_IDS = {
    constants.LOGGING_STR: {
        constants.SMOOTHING_STR: core.KeyId.CFG_NAVSPG_USE_PPP,
        constants.RATE_STR: core.KeyId.CFG_RATE_MEAS,
        constants.MEASUREMENTS_STR: core.KeyId.CFG_MSGOUT_UBX_RXM_RAWX_USB,
        constants.EPHEMERIS_STR: core.KeyId.CFG_MSGOUT_UBX_RXM_SFRBX_USB,
        constants.SOLUTION_STR: core.KeyId.CFG_MSGOUT_UBX_NAV_POSECEF_USB
    },
    constants.CONSTELLATIONS_STR:
    {
        constants.BDS_STR: {
            constants.ENABLED_STR: core.KeyId.CFG_SIGNAL_BDS_ENA,
            constants.SIGNALS_STR: {
                "2I": core.KeyId.CFG_SIGNAL_BDS_B1_ENA, 
                "7I": core.KeyId.CFG_SIGNAL_BDS_B2_ENA
            }   
        },
        constants.GAL_STR: {
            constants.ENABLED_STR: core.KeyId.CFG_SIGNAL_GAL_ENA,
            constants.SIGNALS_STR: {
                "1C": core.KeyId.CFG_SIGNAL_GAL_E1_ENA, 
                "7B": core.KeyId.CFG_SIGNAL_GAL_E5B_ENA
            }
        },
        constants.GLO_STR: {
            constants.ENABLED_STR: core.KeyId.CFG_SIGNAL_GLO_ENA,
            constants.SIGNALS_STR: {
                "1C": core.KeyId.CFG_SIGNAL_GLO_L1_ENA, 
                "2C": core.KeyId.CFG_SIGNAL_GLO_L2_ENA
            }
        },
        constants.GPS_STR: {
            constants.ENABLED_STR: core.KeyId.CFG_SIGNAL_GPS_ENA,
            constants.SIGNALS_STR: {
                "1C": core.KeyId.CFG_SIGNAL_GPS_L1CA_ENA, 
                "2L": core.KeyId.CFG_SIGNAL_GPS_L2C_ENA
            }
        },
        constants.QZSS_STR: {
            constants.ENABLED_STR: core.KeyId.CFG_SIGNAL_QZSS_ENA,
            constants.SIGNALS_STR: {
                "1C": core.KeyId.CFG_SIGNAL_QZSS_L1CA_ENA, 
                "2L": core.KeyId.CFG_SIGNAL_QZSS_L2C_ENA
            }
        }
    }
}

# ------------------------------------------------------------------------------




LOGGING_STR = "logging"
CONSTELLATIONS_STR = 'constellations'
SIGNALS_STR = 'signals'
RATE_STR = 'rate'
MEAS_PROP_STR = 'measurement_properties'
SOLUTION_STR = 'solution'
MEASUREMENTS_STR = 'measurements'
EPHEMERIS_STR = 'ephemeris'
ENABLED_STR = "enabled"
SMOOTHING_STR = "smoothing"
FILE_ROTATION_STR = "file_rotation"

GPS_STR = 'GPS'
GAL_STR = 'GAL'
BDS_STR = 'BDS'
QZSS_STR = 'QZSS'
GLO_STR = 'GLO'



import datetime
import json
import os.path
import sys
import serial
import signal
import time
import queue

from roktools import logger, rinex

from .ublox import config as ublox_config
from .ublox import core as ublox_core
from .ublox import constants
from .ublox import helpers
from .ublox import receiver
from .ublox import SERIAL_PORT_STR, BAUD_RATE_STR
from . import OUTPUT_DIR_STR, NAME_STR, FILE_STR, JSON_STR, RATE_STR, MESSAGES_STR

# ------------------------------------------------------------------------------

_record_thread = None

# ------------------------------------------------------------------------------

def config(**kwargs):

    serial_port = kwargs.get(SERIAL_PORT_STR, None)
    conn_config = receiver.detect_connection(serial_port=serial_port)

    serial_port = conn_config[SERIAL_PORT_STR]
    baudrate = conn_config[BAUD_RATE_STR]

    logger.info(f'Attempting to configure device in serial port [ {serial_port},{baudrate} ]')

    with serial.Serial(serial_port, baudrate) as stream:
        
        config_file = kwargs.get(FILE_STR, None)
        json_file = kwargs.get(JSON_STR, None)
        if config_file:
            logger.info(f'Configuring device from file [ {config_file} ]')
            with open(config_file, 'r') as fh:
                ublox_config.set_from_ucenter_file(stream, fh)

        elif json_file:
            logger.info(f'Configuring device from JSON file [ {json_file} ]')
            with open(json_file, 'r') as fh:
                ublox_config.set_from_json_file(stream, fh) 

        else:
            config_dict = {}

            if RATE_STR in kwargs:
                config_dict[constants.RATE_STR] = {
                    constants.MEASUREMENTS_STR : kwargs[RATE_STR],
                    constants.SOLUTION_STR : kwargs[RATE_STR] * 5
                }

            logger.info(f'Configuring device from command line options: {config_dict}')
            ublox_config.set_from_dict(stream, config_dict)



# ------------------------------------------------------------------------------

def detect(serial_port=None, baudrate=None, messages=False):

    config = receiver.detect_config(serial_port=serial_port, baudrate=baudrate, messages=messages)

    if not config:
        return False

    json.dump(config, sys.stdout, indent=2, sort_keys=True)
    sys.stdout.write('\n')
    return True



# ------------------------------------------------------------------------------

def reset(**kwargs):

    serial_port = kwargs.get(SERIAL_PORT_STR, None)
    if not serial_port:
        logger.critical('Need to specify the port to be reset')
        return False
    else:
        logger.info(f'Resetting device in [ {serial_port} ]')
        receiver.reset(serial_port)
        return True

# ------------------------------------------------------------------------------

def record(**kwargs):

    logger.debug('Record arguments ' + str(kwargs))

    slice_period = kwargs.get('slice', rinex.FilePeriod.DAILY)
    file_rotation = rinex.FilePeriod.from_string(slice_period)
    recorder_config = {
        'serial_port': kwargs.get(SERIAL_PORT_STR, None),
        'baudrate': kwargs.get(BAUD_RATE_STR, None),
        'file_rotation': file_rotation,
        'output_dir': kwargs.get(OUTPUT_DIR_STR, '.'),
        'receiver_name': kwargs.get(NAME_STR, 'UBLX')
    }

    global _record_thread
    _record_thread = receiver.Recorder(**recorder_config)

    _record_thread.start()


# ------------------------------------------------------------------------------

def interruption_handler(sig, frame):
    logger.info('You pressed Ctrl+C, gracefully closing files and serial streams')

    global _record_thread
    _record_thread.stop()

    sys.exit(0)

signal.signal(signal.SIGINT, interruption_handler)



import os
import os.path
import shutil
import struct
import tempfile
import time
import threading

import serial

from roktools import logger, rinex
from roktools.time import weektow_to_datetime

from . import helpers
from . import config as ublox_config
from . import core
from . import SERIAL_PORT_STR, BAUD_RATE_STR

def autodetect_serial_port():
    """
    Attempt to find a serial port in which the receiver is connected. 

    :return: If found, a string with the device specification. Otherwise, None
    """

    serial_port = None

    for port in serial.tools.list_ports.comports():
        if 'u-blox GNSS receiver' in port:
            serial_port = port.device
            break

    return serial_port

# ------------------------------------------------------------------------------

def autodetect_baud_rate(serial_port):
   
    BAUD_RATE_LIST = [9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600]

    baudrate = None
    for baud_rate_candidate in BAUD_RATE_LIST:
        with serial.Serial(serial_port, baud_rate_candidate) as ser:
            if ser.isOpen(): ser.close()
            ser.open()
            doc = ser.read(1000)
            ser.close()

            if b'\xb5\x62' in doc:                        
                baudrate = baud_rate_candidate
                break
        
    return baudrate

# ------------------------------------------------------------------------------

def reset(serial_port):

    serial_stream = serial.Serial(serial_port)
    layer = 3
    payload = struct.pack('<BBBBBBBBBBBBB', 255, 255, 0, 0, 0, 0, 0, 0, 255, 255, 0, 0, layer)
    core.send(core.PacketType.CFG_CFG, payload, serial_stream)

    core.wait_for_ack_nak(core.PacketType.CFG_CFG, serial_stream, timeout=core.TIMEOUT_DEF)
    serial_stream.close()

    logger.info(f'Device reset')

    return True

# ------------------------------------------------------------------------------

def submit_config(doc, serial_port=None, baudrate=None):
    """
    Submit a configuration (using the Receiver configuration standard in JSON
    format)
    """

    if not serial_port:
        serial_port = autodetect_serial_port()

    if not baudrate:
        baudrate = autodetect_baud_rate(serial_port)

    try:
        serial_stream = serial.Serial(serial_port, baudrate)
        ublox_config.set_from_dict(serial_stream, doc)
        serial_stream.close()
        
        return True
 
    except serial.SerialException:
        return False


# ------------------------------------------------------------------------------

def detect_config(serial_port=None, baudrate=None, messages=False):

    config = {}

    try:
        config = detect_connection(serial_port=serial_port, baudrate=baudrate)
        serial_port = config.pop(SERIAL_PORT_STR)
        baudrate = config.pop(BAUD_RATE_STR)
    except Exception as e:
        logger.critical('Could not detect connection: {}. Re-run command'.format(str(e)))
        return None

    with serial.Serial(serial_port, baudrate, **config) as serial_stream:

        try:
            config = ublox_config.get(serial_stream)
            config[SERIAL_PORT_STR] = serial_port
            config[BAUD_RATE_STR] = baudrate
        except Exception as e:
            logger.critical(f'Unable to get receiver parameters. Re-launch the command::{str(e)}')
            return None

        if messages:

            packet_types = set()

            try:
                t_start = time.time()

                for packet in core.ublox_packets(serial_stream, core.TIMEOUT_DEF):

                    elapsed_time = time.time() - t_start

                    ptype = packet[0:2]
                    if ptype not in packet_types:
                        packet_types.add(ptype)

                    if (elapsed_time > 10):
                        break
            except TimeoutError:
                pass

            config['messages'] = [core.PacketType(m).name for m in packet_types]

    return config

# ------------------------------------------------------------------------------


def detect_connection(serial_port=None, baudrate=None):

    if not serial_port:
        logger.debug('No serial port specified, attempting to find one')
        serial_port = autodetect_serial_port()

    if not baudrate:
        logger.debug('No baud rate specified, attempting to find one')
        baudrate = autodetect_baud_rate(serial_port)

    stream = serial.Serial(serial_port, baudrate=baudrate, timeout=core.TIMEOUT_DEF)
    logger.debug(f'Opened connection [ {serial_port} ] --> [ {stream} ]')

    config = {
        SERIAL_PORT_STR: serial_port,
        BAUD_RATE_STR: baudrate
    }

    KEY_IDS = [core.KeyId.CFG_UART1_BAUDRATE, core.KeyId.CFG_UART1_STOPBITS, 
                core.KeyId.CFG_UART1_PARITY, core.KeyId.CFG_UART1_DATABITS]

    key_values = core.submit_valget_packet(KEY_IDS, stream)

    key_id = core.KeyId.CFG_UART1_BAUDRATE
    if key_id in key_values:
        value = key_values[key_id]
        config[BAUD_RATE_STR] = core.parse_key_id_value(key_id, value)

    key_id = core.KeyId.CFG_UART1_STOPBITS
    if key_id in key_values:
        value = key_values[key_id]
        stopbits = core.parse_key_id_value(key_id, value)
        if stopbits == 1:
            config['stopbits'] = serial.STOPBITS_ONE
        elif stopbits == 2:
            config['stopbits'] = serial.STOPBITS_ONE_POINT_FIVE
        elif stopbits == 3:
            config['stopbits'] = serial.STOPBITS_TWO
        else:
            raise ValueError('Half bit not supported by pyserial')

    key_id = core.KeyId.CFG_UART1_PARITY
    if key_id in key_values:
        value = key_values[key_id]
        parity = core.parse_key_id_value(key_id, value)
        if parity == 0:
            config['parity'] = serial.PARITY_NONE
        elif parity == 1:
            config['parity'] = serial.PARITY_ODD
        elif parity == 2:
            config['parity'] = serial.PARITY_EVEN

    key_id = core.KeyId.CFG_UART1_DATABITS
    if key_id in key_values:
        value = key_values[key_id]
        bytesize = core.parse_key_id_value(key_id, value)
        if bytesize == 0:
            config['bytesize'] = serial.EIGHTBITS
        elif bytesize == 1:
            config['bytesize'] = serial.SEVENBITS
    
    logger.debug(f'Serial connection configuration {config}')

    stream.close()

    return config

# ------------------------------------------------------------------------------

class Recorder(threading.Thread):


    def __init__(self, serial_port=None, baudrate=None, file_rotation=rinex.FilePeriod.DAILY,
                 output_dir=".", receiver_name="UBLX"):
        threading.Thread.__init__(self)

        self.slice = file_rotation

        self.output_dir = os.path.abspath(output_dir)
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        self.serial_port = serial_port
        self.baudrate = None
        if not self.serial_port:
            logger.info('Serial port not defined, autodetecting')
            connection_config = detect_connection()
            self.serial_port = connection_config.pop(SERIAL_PORT_STR)
            self.baudrate = connection_config.pop(BAUD_RATE_STR)

        if baudrate:
            self.baudrate = baudrate
        if not self.baudrate:
            self.baudrate = autodetect_baud_rate(self.serial_port)

        self.serial_stream = serial.Serial(self.serial_port, self.baudrate)

        if not self.serial_stream.is_open:
            logger.critical(f'Could not open serial stream [ {self.serial_port} ]')

        self.receiver_name = receiver_name

        self.tempdir = tempfile.mkdtemp(prefix="pyublox_")

        self.current_epoch_suffix = None

        self.fout = None

        logger.info("Writing ubx data from _serial_stream [ {} ] for receiver [ {} ] to files "
                    "in folder [ {} ], with [ {} ] periodicity".format(
                           self.serial_port, self.receiver_name, self.output_dir, self.slice.name))

        logger.debug('Partial files will be written in this temporary folder [ {} ]'.format(self.tempdir))

    # ---

    def run(self):

        logger.debug('Starting recording thread')

        incoming_epoch_suffix = None
        for packet in core.ublox_packets(self.serial_stream, core.TIMEOUT_DEF):

            try:
                packet_type, parsed_packet = core.parse(packet)
            except KeyError:
                continue
            
            if packet_type == core.PacketType.RXM_RAWX:
                tow = parsed_packet.rcvTow
                week = parsed_packet.week
                epoch = weektow_to_datetime(tow, week)

                incoming_epoch_suffix = self.slice.build_rinex3_epoch(epoch)
                logger.debug("U-blox RXM_RAWX packet found! Epoch suffix {}".format(incoming_epoch_suffix))

            if not incoming_epoch_suffix:
                continue

            if not self.current_epoch_suffix or self.current_epoch_suffix != incoming_epoch_suffix:

                self.__close_and_save_partial_file__()

                self.current_epoch_suffix = incoming_epoch_suffix

                filename = os.path.join(self.tempdir,'{}_{}.ubx'.format(self.receiver_name, self.current_epoch_suffix))
                self.fout = open(filename, 'wb')
                logger.info("Created new data file [ {} ]".format(os.path.basename(filename)))

            self.fout.write(core.PREAMBLE)
            self.fout.write(packet)

    # ---

    def stop(self):

        if self.serial_stream:
            self.serial_stream.close()
            logger.info('Closed serial stream from [ {} ]'.format(self.serial_port))

        self.__close_and_save_partial_file__()

        if self.tempdir:
            shutil.rmtree(self.tempdir)

    # ---

    def __close_and_save_partial_file__(self):

        if self.fout:
            filename = self.fout.name

            self.fout.close()

            src = filename
            dst = os.path.join(self.output_dir, os.path.basename(filename))

            logger.debug('Moving [ {} ] -> [ {} ]'.format(src, dst))

            shutil.move(src, dst)



#!/usr/bin/env python
"""
Python client to interface with Ublox receivers

Usage:
    pyublox -h | --help
    pyublox --version
    pyublox detect [ -p <serial-port>] [-b <baudrate>] [-m] [-d <level>]
    pyublox config [ -p <serial-port>] [-b <baudrate>] [-f <file>] [-j <file>]
                   [ -r <rate> ] [ --obs-data ] [ --nav-data ] [ --llh-data ] [-d <level>]
    pyublox record [ -p <serial-port>] [-b <baudrate>] [ -n <name> ] [-o <dir>] 
                   [-s <periodicity>] [-d <level>]
    pyublox reset  [-d <level>] [-b <baudrate>] <serial-port>

Options:
    -h --help         Show this help
    -v --version      Show the version
    -d --debug (DEBUG | INFO | WARNING | CRITICAL)
                      Issue debug information (verbosity controlled by the given debug level) [default: CRITICAL]
    -p --serial-port <serial-port>
                      Serial port to work with (e.g. /dev/ttyACM0)
    -b --baudrate <baudrate> Baud rate of the port (e.g. 115200)
    -f --file <file>  Configure the receiver using a configuration file obtained 
                      with ublox u-center tool.
    -j --json <file>  Configure the receiver using a JSON file with the configuration
                      parameters. A sample of the format for the JSON file can 
                      be obtained with the `detect` command
    -m --messages     When detecting the connection, take a sample and check 
                      which messages are reported
    -r --rate <rate>  Data rate at which the observable will be recorded
    --obs-data        Switch to turn on recording of observables (GNSS measurements)
    --nav-data        Switch to turn on recording of navigation data
    --llh-data        Switch to turn on recording of position estimates
    -o --output-dir <dir>
                      Record the ublox stream to a file or to stdout. If a file 
                      is specified, the periodicity  [default: .]
    -n --name <name>  Specify the receiver name, that will be used to name the 
                      receiver [default: UBLX]
    -s --slice (daily | hourly | quarterly)
                      Slice the data in chunks of given periodicity [default: daily]

Commands:
    detect  Detect ublox device attached to serial port and output basic configuration
    config  Submit configuration options to ublox
    record  Record valid binary UBX packets received from the given port to a file
    reset   Reset the receiver (use when the receiver seems to be hanged or not responsive)
"""
import docopt
import pkg_resources
import sys

from . import commands
from . import OUTPUT_DIR_STR, SLICE_STR, NAME_STR, FILE_STR, JSON_STR, RATE_STR, MESSAGES_STR

from .ublox import SERIAL_PORT_STR, BAUD_RATE_STR

from roktools import logger

def main():
    """
    """

    version = pkg_resources.require("pyublox")[0].version

    args = docopt.docopt(__doc__, version=version, options_first=False)

    logger.set_level(args['--debug'])
    logger.debug(f"Start main, parsed arg\n {args}")

    command, command_args = __get_command__(args)
    logger.debug(f'Command arguments\n {command_args}')


    try:
        command(**command_args)
    except Exception as e:
        sys.stderr.write("FATAL: " + str(e))

    return 0


def __get_command__(args):

    command = None
    command_args = {}

    if args['config']:
        command = commands.config
        command_args = __build_command_args__([SERIAL_PORT_STR, BAUD_RATE_STR, RATE_STR, FILE_STR, JSON_STR], args)

    elif args['detect']:
        command = commands.detect
        command_args = __build_command_args__([SERIAL_PORT_STR, BAUD_RATE_STR, MESSAGES_STR], args)

    elif args['reset']:
        command = commands.reset
        command_args = {
            SERIAL_PORT_STR : args.get(f'<{SERIAL_PORT_STR}>', None),
            BAUD_RATE_STR : args[f'<{BAUD_RATE_STR}>']
        }

    elif args['record']:
        command = commands.record
        command_args = __build_command_args__([SERIAL_PORT_STR, OUTPUT_DIR_STR, SLICE_STR, NAME_STR], args)

    return command, command_args


def __build_command_args__(parameters, arguments):

    command_args = {}
    for arg_params in parameters:
        key = f'--{arg_params}'
        if arguments[key]:
            command_args[arg_params] = arguments[key]

    return command_args


if __name__ == "__main__":

    return_code = main()
    sys.exit(return_code)



from setuptools import find_packages, setup


# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='pyublox',
    version_cc='{version}',
    author='Rokubun',
    author_email='info@rokubun.cat',
    description='Interface with ublox chipsets using Python.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='EUROPEAN SPACE AGENCY PUBLIC LICENSE – V2.3 – STRONG COPYLEFT',
    url="https://gitlab.esa.int/AMIC/pyublox.git",
    setup_requires=['setuptools-git-version-cc'],
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        "docopt",
        "pyserial",
        "roktools>=1.4"
    ],
    entry_points={
        'console_scripts': [
            'pyublox = pyublox.main:main'
        ]
    }
)

